# This file is part of apache-airflow-providers-nomad which is
# released under Apache License 2.0. See file LICENSE or go to
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# for full license details.
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from airflow.exceptions import AirflowException


class NomadProviderException(AirflowException):
    pass


class NomadValidationError(NomadProviderException):
    """Used when can't parse Nomad job input."""


class NomadOperatorError(NomadProviderException):
    """Errors for NomadJobOperator"""


class NomadJobOperatorError(NomadOperatorError):
    """Errors for NomadJobOperator"""


class NomadTaskOperatorError(NomadOperatorError):
    """Errors for NomadJobOperator"""



# This file is part of apache-airflow-providers-nomad which is
# released under Apache License 2.0. See file LICENSE or go to
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# for full license details.
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Executor Interface.
"""

from __future__ import annotations

import argparse
import contextlib
import multiprocessing
from collections.abc import Sequence
from queue import Empty, Queue
from typing import Any

from airflow.executors import workloads
from airflow.executors.base_executor import BaseExecutor
from airflow.executors.workloads import All, ExecuteTask
from airflow.models.taskinstance import TaskInstance
from airflow.models.taskinstancekey import TaskInstanceKey
from airflow.providers.standard.version_compat import AIRFLOW_V_3_0_PLUS
from airflow.utils.state import TaskInstanceState
from sqlalchemy.orm import Session  # type: ignore[import-untyped]

# TaskInstanceKey, Workload in a list, ExecutorConfig
Job = tuple[TaskInstanceKey, list[Any], Any]


class ExecutorInterface(BaseExecutor):
    """Executor with run-queues."""

    supports_ad_hoc_ti_run: bool = True
    serve_logs = True

    def __init__(self, parallelism: int = 1):
        if not AIRFLOW_V_3_0_PLUS:
            raise RuntimeError("Nomad executor only available for Airflow > 3.0.0")

        self._manager = multiprocessing.Manager()
        self.task_queue: Queue[Job] = self._manager.Queue()
        self.parallelism = parallelism
        super().__init__(parallelism=self.parallelism)

    def is_exec_task(self, cmd: Any) -> bool:
        """Safety method in case running in a legacy environment"""
        if isinstance(cmd, ExecuteTask):
            return True
        self.log.error("Workload of unsupported type: '%s'", cmd)
        return False

    def queue_workload(self, workload: All, session: Session) -> None:
        if not self.is_exec_task(workload):
            return
        ti = workload.ti
        self.queued_tasks[ti.key] = workload  # type: ignore

    def _process_workloads(self, workloads: Sequence[workloads.All]) -> None:
        for w in workloads:
            if not self.is_exec_task(w) or not w.ti:
                return

            # TODO: AIP-72 handle populating tokens once https://github.com/apache/airflow/issues/45107 is handled.
            command = [w]
            key = w.ti.key
            queue = w.ti.queue
            executor_config = w.ti.executor_config or {}

            del self.queued_tasks[key]
            self.execute_async(
                key=key, command=command, queue=queue, executor_config=executor_config
            )
            self.running.add(key)

    def execute_async(
        self,
        key: TaskInstanceKey,
        command: Any,
        queue: str | None = None,
        executor_config: Any | None = None,
    ) -> None:
        """Execute task asynchronously."""
        self.log.info("Adding task %s with command %s to run queue", key, command)
        self.event_buffer[key] = (TaskInstanceState.QUEUED, None)
        self.task_queue.put((key, command, executor_config))

    def set_state(self, key: TaskInstanceKey, state: TaskInstanceState, info: str = "") -> None:
        self.log.info("Setting state for key: %s state: %s info: %s", key, state, info)

        if state not in TaskInstanceState:
            self.log.error("Unknown task state %s, setting it as failed", state)
            self.fail(key)
            return

        if state == TaskInstanceState.QUEUED:
            self.queued(key)
        elif state == TaskInstanceState.SUCCESS:
            self.success(key, info=info)
        elif state == TaskInstanceState.FAILED:
            self.fail(key, info=info)
        elif state not in [
            TaskInstanceState.RUNNING,
            TaskInstanceState.RESTARTING,
        ]:
            self.change_state(key, state, remove_running=True, info=info)  # type: ignore[reportArgumentType]

    def sync(self) -> None:
        """Synchronize task state."""
        self.log.debug("self.running: %s", self.running)
        self.log.debug("self.queued: %s", self.queued_tasks)

        for ti_key in list(self.running | set(self.queued_tasks)):
            if status := self.remove_job_if_hanging(ti_key):
                state, info = status
                self.set_state(ti_key, state=state, info=info)

        with contextlib.suppress(Empty):
            while (task := self.task_queue.get_nowait()) and self.slots_available > 0:
                try:
                    key, command, _ = task
                except Exception as err:
                    self.log.exception(
                        "Failed to run task (task/command could be retrieved: (%s))",
                        str(err),
                    )
                    return

                try:
                    self.run_task(task)
                except Exception as err:
                    self.log.exception(
                        "Failed to run task %s with command %s (%s)", key, command, str(err)
                    )
                finally:
                    self.task_queue.task_done()

    def workload_to_command_args(self, workload: ExecuteTask) -> list[str]:
        """Convert a workload object to Task SDK command arguments."""
        return [workload.model_dump_json()]

    def _not_exec_task(self, cmd: Any) -> str | None:
        """Safety method in case running in a legacy environment"""
        if (isinstance(cmd, list) and len(cmd) > 0) and isinstance(cmd[0], ExecuteTask):
            return None
        msg = f"Workload of unsupported type: {cmd}"
        self.log.error(msg)
        return msg

    def run_task(self, task: Job) -> None:
        """Run the next task in the queue."""
        key, cmd, _ = task
        dag_id, task_id, run_id, try_number, _ = key
        if not (isinstance(cmd, list) and len(cmd) > 0) or not self.is_exec_task(cmd[0]):
            self.fail(key, info=f"Unsupported workload '{cmd}'")
            return

        command = self.workload_to_command_args(cmd[0])

        self.log.info(
            "Runing task (%s) from dag (%s) with run ID (%s) (retries %s)",
            task_id,
            dag_id,
            run_id,
            try_number,
        )

        job_template = self.prepare_job_template(key, command)
        try:
            failed_info = self.run_job(job_template)
        except Exception as err:
            failed_info = f"Couldn't run job {key}: {err}"

        if failed_info:
            self.log.error(failed_info)
            self.fail(key, info=failed_info)

    def prepare_job_template(self, key: TaskInstanceKey, command: list[str]) -> dict[str, Any]:
        """Adjutst template to suit upcoming job execution

        :param key: reference to the task instance in question
        :return: job template as as dictionary
        """
        self.log.debug(f"Executing key {key} with command {command}")
        raise NotImplementedError

    def run_job(self, job_template: dict[str, Any] | None) -> str | None:
        """Execute the job defined by a potential job template

        :param: Job template corresponding to the job
        :return: No news is good news, or the error that occured on execution attempt
        """
        self.log.debug(f"Executing template {job_template}")
        raise NotImplementedError

    def remove_job_if_hanging(self, key: TaskInstanceKey) -> tuple[TaskInstanceState, str] | None:
        """Whether the job failed (typically outside of Airflow execution)

        :param key: reference to the task instance in question
        :return: either a tuple of: True/False, potential task status to set (typically: FAILED), additional info
                 or None if no data could be retrieved for the job
        """
        self.log.debug(f"Evaluating status of key {key}")
        raise NotImplementedError

    def get_task_log(self, ti: TaskInstance, try_number: int) -> tuple[list[str], list[str]]:
        self.log.debug(f"Retrieving logs for {ti.key} with {try_number}")
        raise NotImplementedError

    def end(self) -> None:
        """Shut down the executor."""
        self.log.info("Shutting down Nomad executor")
        try:
            # self.log.debug("Flushing task_queue...")
            # self._flush_task_queue()
            # Both queues should be empty...
            self.task_queue.join()
        except ConnectionResetError:
            self.log.exception("Connection Reset error while flushing task_queue.")
        except Exception:
            self.log.exception("Unknown error while flushing task queue and result queue.")
        self._manager.shutdown()


def _get_parser() -> argparse.ArgumentParser:
    """
    Generate documentation; used by Sphinx.

    :meta private:
    """
    return ExecutorInterface._get_parser()



# This file is part of apache-airflow-providers-nomad which is
# released under Apache License 2.0. See file LICENSE or go to
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# for full license details.
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations

"""Logging module to fetch logs via the Nomad API"""

import copy
import logging

import airflow.logging_config
from airflow.config_templates.airflow_local_settings import DEFAULT_LOGGING_CONFIG

from airflow.providers.nomad.generic_interfaces.executor_log_handlers import ExecutorLogLinesHandler

logger = logging.getLogger(__name__)


class NomadLogHandler(ExecutorLogLinesHandler):
    """Extended handler to retrieve logs directly from Nomad"""

    name = "nomad_log_handler"


NOMAD_HANDLER_NAME = NomadLogHandler.name

NOMAD_LOG_CONFIG = copy.deepcopy(DEFAULT_LOGGING_CONFIG)

NOMAD_LOG_CONFIG["handlers"][NOMAD_HANDLER_NAME] = {
    "class": "airflow.providers.nomad.log.NomadLogHandler",
    "formatter": "airflow",
    "filters": ["mask_secrets"],
}

NOMAD_LOG_CONFIG["loggers"]["airflow.task"]["handlers"].append(NOMAD_HANDLER_NAME)


# Due to bug on loading config for services such as dag-processor
# Reproduce: uncomment these lines and run `airflow dag-processor
airflow.logging_config.REMOTE_TASK_LOG = None



# This file is part of apache-airflow-providers-nomad which is
# released under Apache License 2.0. See file LICENSE or go to
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# for full license details.
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
#
# NOTE! THIS FILE IS AUTOMATICALLY GENERATED AND WILL BE
# OVERWRITTEN WHEN PREPARING DOCUMENTATION FOR THE PACKAGES.
#
# IF YOU WANT TO MODIFY THIS FILE, YOU SHOULD MODIFY THE TEMPLATE
# `PROVIDER__INIT__PY_TEMPLATE.py.jinja2` IN the `dev/breeze/src/airflow_breeze/templates` DIRECTORY
#
from __future__ import annotations

import packaging.version

from airflow import __version__ as airflow_version  # type: ignore

__all__ = ["__version__"]

__version__ = "0.0.5"

if packaging.version.parse(  # pragma: no cover
    packaging.version.parse(airflow_version).base_version
) < packaging.version.parse("3.0.0"):
    raise RuntimeError(
        f"The package `apache-airflow-providers-nomad:{__version__}` needs Apache Airflow 3.0.0+"
    )



# This file is part of apache-airflow-providers-nomad which is
# released under Apache License 2.0. See file LICENSE or go to
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# for full license details.
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from collections.abc import Collection

from airflow.sdk import Context
from pydantic import ValidationError

from airflow.providers.nomad.generic_interfaces.nomad_operator_interface import NomadOperator
from airflow.providers.nomad.models import NomadJobModel
from airflow.providers.nomad.templates.job_template import default_task_template


class NomadJobOperator(NomadOperator):
    """Nomad Operator for direct template submission"""

    template_fields: Collection[str] = ["template_content", "template_path"]

    def __init__(
        self,
        template_path: str | None = None,
        template_content: str | None = None,
        observe: bool = True,
        job_log_file: str | None = None,
        **kwargs,
    ):
        if template_path and template_content:
            raise ValueError("Only one of 'template_content' and 'template_path' can be specified")
        self.template_content = template_content
        self.template_path = template_path
        super().__init__(observe=observe, job_log_file=job_log_file, **kwargs)

    def prepare_job_template(self, context: Context):
        if self.template_content:
            self.template = self.nomad_mgr.parse_template_content(self.template_content)
            return

        content = None
        if self.template_path:
            filepath = self.figure_path(self.template_path)
            try:
                with open(filepath) as f:
                    content = f.read()
            except (OSError, IOError) as err:
                self.log.error(f"Can't load job template ({err})")
                return

        if content or (content := context.get("params", {}).get("template_content", "")):
            self.template = self.nomad_mgr.parse_template_content(content)
            return

        try:
            self.template = NomadJobModel.model_validate(default_task_template)
        except ValidationError:
            self.log.error("Default template validation failed")



# This file is part of apache-airflow-providers-nomad which is
# released under Apache License 2.0. See file LICENSE or go to
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# for full license details.
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from enum import Enum
from typing import Any, TypeAlias
import logging

from pydantic import BaseModel, ConfigDict, TypeAdapter, model_validator
from typing_extensions import Self


logger = logging.getLogger(__name__)


class JobType(str, Enum):
    batch = "batch"
    service = "service"


class JobInfoStatus(str, Enum):
    dead = "dead"
    pending = "pending"
    running = "running"


class JobEvalStatus(str, Enum):
    complete = "complete"
    blocked = "blocked"
    pending = "pending"
    canceled = "canceled"
    failed = "failed"

    @classmethod
    def done_states(cls) -> list["JobEvalStatus"]:
        return [cls.failed, cls.canceled, cls.complete]


class Resource(BaseModel):
    model_config = ConfigDict(extra="allow", validate_assignment=True)

    CPU: int | None = None
    MemoryMB: int | None = 256
    DiskMB: int | None = None
    Cores: int | None = None
    MemoryMaxMB: int | None = None
    Networks: list[Any] | None = None
    Devices: list[Any] | None = None
    MBits: int | None = None

    @model_validator(mode="after")
    def compatible_fields(self) -> Self:
        if self.CPU and self.Cores:
            raise ValueError("Both CPU and Cores can't be specified")
        if self.DiskMB and self.DiskMB > 0:
            logger.warning(
                "According to https://github.com/hashicorp/nomad/issues/10009#issuecomment-777445224,"
                " DiskMB is not supported"
            )
            self.DiskMB = None
        return self


class NomadEphemeralDisk(BaseModel):
    Migrate: bool | None = None
    SizeMB: int | None = None
    Sticky: bool | None = None


class TaskConfig(BaseModel):
    model_config = ConfigDict(extra="allow", validate_assignment=True)

    image: str | None = None
    entrypoint: list[str] | None = None
    args: list[str] | None = None
    command: str | list[str] | None = None

    @model_validator(mode="after")
    def compatible_fields(self) -> Self:
        if self.entrypoint and self.command:
            raise ValueError("Both 'entrypoint' and 'command' specified")
        return self


class Volume(BaseModel):
    model_config = ConfigDict(extra="allow", validate_assignment=True)

    AccessMode: str = ""
    AttachmentMode: str = ""
    MountOptions: Any | None = None
    Name: str
    PerAlloc: bool
    ReadOnly: bool
    Source: str
    Sticky: bool
    Type: str


NomadVolList: TypeAlias = dict[str, Volume]
NomadVolumes = TypeAdapter(NomadVolList)


class VolumeMount(BaseModel):
    model_config = ConfigDict(extra="allow", validate_assignment=True)

    Destination: str
    PropagationMode: str
    ReadOnly: bool
    SELinuxLabel: str = ""
    Volume: str

    def from_volumes(self, volumes: NomadVolList) -> bool:
        return self.Volume in volumes


NomadVolMntList: TypeAlias = list[VolumeMount]
NomadVolumeMounts = TypeAdapter(NomadVolMntList)


class Task(BaseModel):
    model_config = ConfigDict(extra="allow", validate_assignment=True)

    Config: TaskConfig
    Name: str
    Resources: Resource | None = None
    Driver: str
    Env: dict[str, str] | None = None
    VolumeMounts: NomadVolMntList | None = None


class TaskGroup(BaseModel):
    model_config = ConfigDict(extra="allow", validate_assignment=True)

    Tasks: list[Task]
    Name: str
    Count: int | None = None
    Volumes: NomadVolList | None = None
    EphemeralDisk: NomadEphemeralDisk | None = None


class Job(BaseModel):
    model_config = ConfigDict(extra="allow", validate_assignment=True)
    TaskGroups: list[TaskGroup]
    ID: str
    Name: str
    Namespace: str | None = None
    Type: JobType


class NomadJobModel(BaseModel):
    model_config = ConfigDict(extra="allow", validate_assignment=True)
    Job: Job

    def tasknames(self) -> list[str]:
        return [task.Name for group in self.Job.TaskGroups for task in group.Tasks]


class NomadFailedAllocInfo(BaseModel):
    AllocationTime: int
    ClassExhausted: Any | None = None
    ClassFiltered: Any | None = None
    CoalescedFailures: int
    ConstraintFiltered: dict[str, int] | None = None
    DimensionExhausted: dict[str, int] | None = None
    NodePool: str
    NodesAvailable: dict[str, int]
    NodesEvaluated: int
    NodesExhausted: int
    NodesFiltered: int
    NodesInPool: int
    QuotaExhausted: Any | None = None
    ResourcesExhausted: dict[str, dict] | None = None
    ScoreMetaData: Any | None = None
    Scores: Any | None = None

    def errors(self) -> list[Any]:
        """Turn an evaluation failure record into an error message"""
        errors: set[Any] = set()
        if self.ClassExhausted:
            errors.add(str(self.ClassExhausted))
        if self.ClassFiltered:
            errors.add(str(self.ClassFiltered))
        if self.ConstraintFiltered:
            errors.add(str(self.ConstraintFiltered))
        if self.DimensionExhausted:
            errors.add(str(self.DimensionExhausted))
        if self.QuotaExhausted:
            errors.add(str(self.QuotaExhausted))
        if self.ResourcesExhausted:
            errors.add(str(self.ResourcesExhausted))
        return list(errors)


class NomadJobEvaluationInfo(BaseModel):
    BlockedEval: str | None = None
    ClassEligibility: dict[str, bool] | None = None
    CreateIndex: int
    CreateTime: int
    FailedTGAllocs: dict[str, NomadFailedAllocInfo] | None = None
    ID: str
    JobID: str
    JobModifyIndex: int | None = None
    ModifyIndex: int
    ModifyTime: int
    Namespace: str
    PreviousEval: str | None = None
    Priority: int
    SnapshotIndex: int | None = None
    Status: JobEvalStatus
    StatusDescription: str | None = None
    TriggeredBy: str
    Type: JobType


NomadJobEvalList: TypeAlias = list[NomadJobEvaluationInfo]
NomadJobEvaluation = TypeAdapter(NomadJobEvalList)


class NomadJobSubmission(BaseModel):
    model_config = ConfigDict(extra="allow", validate_assignment=True)

    Status: JobInfoStatus
    Namespace: str
    ID: str
    Name: str
    Type: JobType
    Priority: int
    AllAtOnce: bool
    Datacenters: list[str]
    NodePool: str
    TaskGroups: list[TaskGroup]


class NomadEvent(BaseModel):
    Details: dict
    DiskLimit: int
    DisplayMessage: str
    DownloadError: str
    DriverError: str
    DriverMessage: str
    ExitCode: int
    FailedSibling: str
    FailsTask: bool
    GenericSource: str
    KillError: str
    KillReason: str
    KillTimeout: int
    Message: str
    RestartReason: str
    SetupError: str
    Signal: int
    StartDelay: int
    TaskSignal: str
    TaskSignalReason: str
    Time: int
    Type: str
    ValidationError: str
    VaultError: str


class NomadTaskState(BaseModel):
    model_config = ConfigDict(extra="allow", validate_assignment=True)

    Events: list[NomadEvent]
    Failed: bool
    State: str


class NomadJobAllocationInfo(BaseModel):
    model_config = ConfigDict(extra="allow", validate_assignment=True)

    ClientStatus: str
    EvalID: str
    FollowupEvalID: str
    ID: str
    JobID: str
    JobType: JobType
    JobVersion: int
    Name: str
    Namespace: str
    NextAllocation: str
    NodeID: str
    TaskGroup: str
    TaskStates: dict[str, NomadTaskState]

    def errors(self) -> dict[str, dict[str, list[Any]]]:
        """Turn an evaluation failure record into an error message"""
        errors: dict[str, dict[str, list[Any]]] = {}
        for task in self.TaskStates:
            errors[task] = {}
            for event in self.TaskStates[task].Events:
                errors[task][event.Type] = []
                if event.DownloadError:
                    errors[task][event.Type].append(event.DownloadError)
                if event.DriverError:
                    errors[task][event.Type].append(event.DriverError)
                if event.DriverError:
                    errors[task][event.Type].append(event.DriverError)
                if event.KillError:
                    errors[task][event.Type].append(event.KillError)
                if event.KillReason:
                    errors[task][event.Type].append(event.KillReason)
                if event.ValidationError:
                    errors[task][event.Type].append(event.ValidationError)
                if event.VaultError:
                    errors[task][event.Type].append(event.VaultError)
                if not errors[task][event.Type]:
                    errors[task].pop(event.Type)
            if not errors[task]:
                errors.pop(task)
        return errors


NomadJobAllocList: TypeAlias = list[NomadJobAllocationInfo]
NomadJobAllocations = TypeAdapter(NomadJobAllocList)


class NomadChildrenSummary(BaseModel):
    Dead: int
    Pending: int
    Running: int


class NomadJobSummaryInfo(BaseModel):
    Complete: int
    Failed: int
    Lost: int
    Queued: int
    Running: int
    Starting: int
    Unknown: int


class NomadJobSummary(BaseModel):
    model_config = ConfigDict(extra="allow", validate_assignment=True)

    Children: NomadChildrenSummary
    JobID: str
    Namespace: str
    Summary: dict[str, NomadJobSummaryInfo]

    def all_failed(self) -> bool:
        for taskgroup in self.Summary.values():
            if not (
                (taskgroup.Failed > 0 or taskgroup.Lost > 0 or taskgroup.Unknown > 0)
                and taskgroup.Complete == 0
                and taskgroup.Queued == 0
                and taskgroup.Running == 0
                and taskgroup.Starting == 0
            ):
                return False
        return True

    def all_done(self) -> bool:
        for taskgroup in self.Summary.values():
            if not (
                (taskgroup.Failed > 0 or taskgroup.Lost > 0 or taskgroup.Complete > 0)
                and taskgroup.Queued == 0
                and taskgroup.Running == 0
                and taskgroup.Starting == 0
                and taskgroup.Unknown == 0
            ):
                return False
        return True



# This file is part of apache-airflow-providers-nomad which is
# released under Apache License 2.0. See file LICENSE or go to
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# for full license details.
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import time
from pathlib import Path

from airflow.configuration import conf
from airflow.models import BaseOperator
from airflow.sdk import Context
from nomad.api.exceptions import BaseNomadException  # type: ignore[import-untyped]

from airflow.providers.nomad.constants import CONFIG_SECTION
from airflow.providers.nomad.exceptions import NomadOperatorError
from airflow.providers.nomad.manager import NomadManager
from airflow.providers.nomad.models import (
    JobEvalStatus,
    JobInfoStatus,
    NomadJobAllocList,
    NomadJobModel,
)


class NomadOperator(BaseOperator):
    """Abstract class to provide shared functionality across Nomad Operators"""

    def __init__(self, observe: bool = True, job_log_file: str | None = None, **kwargs):
        self.nomad_mgr = NomadManager()
        self.nomad_mgr.initialize()
        self.observe = observe
        self.job_log_file = job_log_file
        self.operator_poll_delay: int = conf.getint(
            CONFIG_SECTION, "operator_poll_delay", fallback=10
        )
        self.template: NomadJobModel | None = None
        super().__init__(**kwargs)

    @staticmethod
    def sanitize_logs(alloc_id: str, task_name: str, logs: str) -> str:
        if not logs:
            return logs

        sanitized_logs = logs
        fileloc = Path(f"{alloc_id}-{task_name}.log")
        if fileloc.is_file():
            with fileloc.open("r") as file:
                prefix = file.read()
                sanitized_logs = logs[len(prefix) :]

        with fileloc.open("w") as file:
            file.write(logs)

        return sanitized_logs

    def get_children_output(
        self, job_id, alloc_info: NomadJobAllocList | None = None
    ) -> tuple[str, str]:
        if not alloc_info and not (alloc_info := self.nomad_mgr.get_nomad_job_allocation(job_id)):
            return "", ""

        all_logs = ""
        all_output = []
        for allocation in alloc_info:
            for task_name in allocation.TaskStates:
                logs = ""
                if self.job_log_file:
                    logs += self.nomad_mgr.get_job_file(allocation.ID, self.job_log_file)

                if output := self.nomad_mgr.get_job_stdout(allocation.ID, task_name):
                    all_output.append(output.strip())
                    logs += output

                logs += self.nomad_mgr.get_job_stderr(allocation.ID, task_name)

                all_logs += self.sanitize_logs(allocation.ID, task_name, logs)

                for line in logs.splitlines():
                    self.log.info("[job: %s][alloc: %s] %s", job_id, allocation.ID, line)

        return ",".join(all_output), all_logs

    @staticmethod
    def figure_path(path_str: str):
        path = Path(path_str)
        if not path.is_absolute():
            path = Path(conf.get_mandatory_value("core", "dags_folder")) / path
        return path

    def prepare_job_template(self, context: Context):
        raise NotImplementedError

    def execute(self, context: Context):
        if not self.template:
            self.prepare_job_template(context)

        if not self.template:
            raise NomadOperatorError("Nothing to execute")

        job_id = self.template.Job.ID
        try:
            response = self.nomad_mgr.nomad.job.register_job(  # type: ignore[optionalMemberAccess, union-attr]
                job_id, self.template.model_dump(exclude_unset=True)
            )
        except BaseNomadException as err:
            raise NomadOperatorError(
                f"Job submission failed ({err}), job template: {self.template.model_dump_json()}"
            )

        if not response:
            self.log.warning("No response on job submission attempt, may suggest an error")

        if not self.observe:
            return

        status: JobInfoStatus | None = JobInfoStatus.pending
        job_info: list[str] = []
        logs = ""
        job_status, job_alloc, job_eval, job_summary = None, None, None, None
        job_snapshot = {}
        output = ""
        all_done = False
        while status != JobInfoStatus.dead and not all_done:
            job_info = []

            # Snapshotting state
            job_status = self.nomad_mgr.get_nomad_job_submission(job_id)
            job_alloc = self.nomad_mgr.get_nomad_job_allocation(job_id)
            job_eval = self.nomad_mgr.get_nomad_job_evaluations(job_id)
            job_summary = self.nomad_mgr.get_nomad_job_summary(job_id)
            job_snapshot = {
                "job_summary": job_summary,
                "job_alloc": job_alloc,
                "job_eval": job_eval,
            }

            output, logs = self.get_children_output(job_id, alloc_info=job_alloc)

            if not job_status or job_status.Status != JobInfoStatus.running:
                if (
                    result := self.nomad_mgr.remove_job_if_hanging(
                        job_id, ignore_dead=True, job_status=job_status, **job_snapshot
                    )
                ) and result[0]:
                    _, error = result
                    if job_info := self.nomad_mgr.job_all_info_str(job_id, **job_snapshot):
                        raise NomadOperatorError(
                            f"Job {job_id} got killed due to error: {error}\n"
                            "Additional info:\n"
                            "\n".join(job_info)
                        )
                    break

            time.sleep(self.operator_poll_delay)
            status = job_status.Status if job_status else None
            all_done = job_summary.all_done() if job_summary else False

        # Collecting final status
        job_alloc = self.nomad_mgr.get_nomad_job_allocation(job_id)
        job_eval = self.nomad_mgr.get_nomad_job_evaluations(job_id)
        job_summary = self.nomad_mgr.get_nomad_job_summary(job_id)
        job_info = self.nomad_mgr.job_all_info_str(
            job_id, job_alloc=job_alloc, job_eval=job_eval, job_summary=job_summary
        )
        job_info_str = "\n".join(job_info)
        if job_eval and any(evalu.Status == JobEvalStatus.complete for evalu in job_eval):
            if output:
                return output
            if job_summary and not job_summary.all_failed():
                return (
                    f"No output from job. Logs/stderr: {logs.splitlines()}\nJob info: \n"
                    + "\n".join(job_info)
                )

        raise NomadOperatorError(f"Job submission failed {job_info_str}")



# This file is part of apache-airflow-providers-nomad which is
# released under Apache License 2.0. See file LICENSE or go to
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# for full license details.
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

CONFIG_SECTION = "nomad_provider"



# This file is part of apache-airflow-providers-nomad which is
# released under Apache License 2.0. See file LICENSE or go to
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# for full license details.
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import json
import logging
from json.decoder import JSONDecodeError
from pathlib import Path
from typing import Any

from airflow.models.taskinstance import TaskInstance
from airflow.models.taskinstancekey import TaskInstanceKey
from airflow.sdk.types import RuntimeTaskInstanceProtocol
from pydantic import ValidationError

from airflow.providers.nomad.exceptions import NomadValidationError
from airflow.providers.nomad.models import NomadJobModel

logger = logging.getLogger(__name__)


def job_id(dag_id: str, task_id: str, run_id: str, try_number: int, map_index: int):
    return f"{dag_id}-{task_id}-{run_id}-{try_number}-{map_index}"


def job_id_from_taskinstance_key(key: TaskInstanceKey) -> str:
    dag_id, task_id, run_id, try_number, map_index = key
    return job_id(dag_id, task_id, run_id, try_number, map_index)


def job_id_from_taskinstance(ti: TaskInstance | RuntimeTaskInstanceProtocol) -> str:
    index = ti.map_index if ti.map_index else -1
    run_id = ti.run_id if ti.run_id else ""
    return job_id(ti.dag_id, ti.task_id, run_id, ti.try_number, index)


def job_task_id_from_taskinstance_key(key: TaskInstanceKey) -> str:
    dag_id, task_id, _, _, _ = key
    return f"{dag_id}-{task_id}"


def validate_nomad_job(data: dict[str, Any]) -> NomadJobModel:
    try:
        return NomadJobModel.model_validate(data)
    except ValidationError as err:
        raise NomadValidationError(err)


def validate_nomad_job_json(json_str: str) -> NomadJobModel:
    try:
        return NomadJobModel.model_validate_json(json_str)
    except (ValidationError, JSONDecodeError) as err:
        raise NomadValidationError(err)


def parse_json_job_template(path: Path) -> NomadJobModel | None:
    """try to parse a json or hcl input as a nomad job template"""
    try:
        file = open(str(path))
    except (IOError, OSError) as err:
        logger.error(f"Can't open file: {path}' ({err})")
        return  # type: ignore [return-value]

    with file:
        try:
            data = json.load(file)
        except json.JSONDecodeError:
            return None
        return validate_nomad_job(data)


def dict_to_lines(d: dict[str, Any]) -> list[str]:
    return json.dumps(d, sort_keys=True, indent=4).splitlines()
